import os


PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BASE_DIR = os.path.dirname(PROJECT_DIR)

DEBUG = True

ALLOWED_HOSTS = ["*"]

SECRET_KEY = "not needed"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "wagtailinventory.sqlite",
    },
}

WAGTAIL_APPS = (
    "wagtail.contrib.forms",
    "wagtail.contrib.settings",
    "wagtail.admin",
    "wagtail.core",
    "wagtail.documents",
    "wagtail.images",
    "wagtail.sites",
    "wagtail.users",
)

WAGTAILADMIN_RICH_TEXT_EDITORS = {
    "default": {"WIDGET": "wagtail.admin.rich_text.DraftailRichTextArea"},
    "custom": {"WIDGET": "wagtail.tests.testapp.rich_text.CustomRichTextArea"},
}

MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)

INSTALLED_APPS = (
    (
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.messages",
        "django.contrib.sessions",
        "django.contrib.staticfiles",
        "taggit",
    )
    + WAGTAIL_APPS
    + (
        "wagtailinventory",
        "wagtailinventory.tests.testapp",
    )
)

STATIC_ROOT = os.path.join(BASE_DIR, "static")
STATIC_URL = "/static/"

MEDIA_ROOT = os.path.join(BASE_DIR, "media")
MEDIA_URL = "/media/"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ],
            "debug": True,
        },
    },
]

WAGTAIL_SITE_NAME = "Test Site"

ROOT_URLCONF = "wagtailinventory.tests.urls"



# -*- coding: utf-8 -*-
from django.db import migrations, models
import django.db.models.deletion

from wagtail.core import blocks as core_blocks
from wagtail.core import fields as core_fields  # pragma: no cover


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0032_add_bulk_delete_page_permission'),
    ]

    operations = [
        migrations.CreateModel(
            name='MultipleStreamFieldsPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('first', core_fields.StreamField([('atom', core_blocks.StructBlock([('title', core_blocks.CharBlock())])), ('molecule', core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])), ('organism', core_blocks.StructBlock([('molecules', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])))]))])),
                ('second', core_fields.StreamField([('atom', core_blocks.StructBlock([('title', core_blocks.CharBlock())])), ('molecule', core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])), ('organism', core_blocks.StructBlock([('molecules', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])))]))])),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='NoStreamFieldsPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('content', models.TextField()),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='SingleStreamFieldPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('content', core_fields.StreamField([('text', core_blocks.CharBlock()), ('atom', core_blocks.StructBlock([('title', core_blocks.CharBlock())])), ('molecule', core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])), ('organism', core_blocks.StructBlock([('molecules', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])))]))])),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
    ]



from django.test import TestCase

from wagtail.core.models import Page

from wagtailinventory.models import PageBlock


class TestPageBlock(TestCase):
    def test_page_str(self):
        page_block = PageBlock(
            page=Page(title="Title", slug="title"), block="path.to.block"
        )
        self.assertEqual(str(page_block), "<Title, path.to.block>")



from wagtail.core import blocks


class Atom(blocks.StructBlock):
    title = blocks.CharBlock()


class Molecule(blocks.StructBlock):
    title = blocks.CharBlock()
    atoms = blocks.ListBlock(Atom())


class Organism(blocks.StructBlock):
    molecules = blocks.ListBlock(Molecule())



from wagtailinventory.views import SearchView


try:
    from django.urls import re_path
except ImportError:
    from django.conf.urls import url as re_path


app_name = "wagtailinventory"


urlpatterns = [
    re_path(r"^$", SearchView.as_view(), name="search"),
]



from django.http import HttpResponseBadRequest
from django.shortcuts import render
from django.views.generic import View

import wagtail
from wagtail.core.models import Page

from wagtailinventory.forms import PageBlockQueryFormSet


class SearchView(View):
    template_name = "wagtailinventory/search.html"

    def get(self, request):
        if len(request.GET) > 1:
            formset = PageBlockQueryFormSet(request.GET)
            if not formset.is_valid():
                return HttpResponseBadRequest("invalid query")

            queryset = formset.get_query()
        else:
            formset = PageBlockQueryFormSet()
            queryset = Page.objects.all()

        queryset = queryset.order_by("title")

        # https://docs.wagtail.io/en/latest/releases/2.5.html#changes-to-admin-pagination-helpers
        if wagtail.VERSION < (2, 5):
            from wagtail.utils.pagination import paginate

            paginator, pages = paginate(request, queryset)
        else:  # pragma: no cover
            from django.core.paginator import Paginator

            paginator = Paginator(queryset, per_page=20)
            pages = paginator.get_page(request.GET.get("p"))

        for page in pages:
            page.can_choose = True

        return render(
            request, self.template_name, {"formset": formset, "pages": pages}
        )



from django.test import TestCase

from wagtail.core.models import Page

from wagtailinventory.helpers import get_page_blocks


CORE_BLOCKS = "wagtail.core.blocks"


class TestGetPageBlocks(TestCase):
    fixtures = ["test_blocks.json"]

    def test_page_with_no_streamfields_returns_empty_list(self):
        page = Page.objects.get(slug="no-streamfields-page")
        self.assertEqual(get_page_blocks(page), [])

    def test_empty_streamfield_returns_empty_list(self):
        page = Page.objects.get(slug="single-streamfield-page-no-content")
        self.assertEqual(get_page_blocks(page), [])

    def test_streamfield_with_single_block(self):
        page = Page.objects.get(slug="single-streamfield-page-content")
        self.assertEqual(
            get_page_blocks(page),
            [
                CORE_BLOCKS + ".field_block.CharBlock",
                "wagtailinventory.tests.testapp.blocks.Atom",
            ],
        )

    def test_multiple_streamfields(self):
        page = Page.objects.get(slug="multiple-streamfields-page")
        self.assertEqual(
            get_page_blocks(page),
            [
                CORE_BLOCKS + ".field_block.CharBlock",
                CORE_BLOCKS + ".list_block.ListBlock",
                "wagtailinventory.tests.testapp.blocks.Atom",
                "wagtailinventory.tests.testapp.blocks.Molecule",
                "wagtailinventory.tests.testapp.blocks.Organism",
            ],
        )

    def test_nested_streamblocks(self):
        page = Page.objects.get(slug="nested-streamblock-page")
        self.assertEqual(
            get_page_blocks(page),
            [
                CORE_BLOCKS + ".field_block.CharBlock",
                CORE_BLOCKS + ".stream_block.StreamBlock",
                "wagtailinventory.tests.testapp.blocks.Atom",
            ],
        )



from django.db import models

from wagtail.core.models import Page


class PageBlock(models.Model):
    page = models.ForeignKey(
        Page, related_name="page_blocks", on_delete=models.CASCADE
    )
    block = models.CharField(max_length=255, db_index=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["page", "block"], name="unique_page_block"
            ),
        ]

    def __str__(self):
        return "<{}, {}>".format(self.page, self.block)



from django.urls import include, re_path, reverse

from wagtail.admin.menu import MenuItem
from wagtail.core import hooks

from wagtailinventory import urls
from wagtailinventory.helpers import (
    create_page_inventory,
    delete_page_inventory,
    update_page_inventory,
)


@hooks.register("after_create_page")
def do_after_page_create(request, page):
    create_page_inventory(page)


@hooks.register("after_edit_page")
def do_after_page_edit(request, page):
    update_page_inventory(page)


@hooks.register("after_delete_page")
def do_after_page_dete(request, page):
    delete_page_inventory(page)


@hooks.register("register_admin_urls")
def register_inventory_urls():
    return [
        re_path(r"^inventory/", include(urls, namespace="wagtailinventory")),
    ]


@hooks.register("register_settings_menu_item")
def register_inventory_menu_item():
    return MenuItem(
        "Block Inventory",
        reverse("wagtailinventory:search"),
        classnames="icon icon-placeholder",
        order=11000,
    )



from django.db import models

from wagtail.admin.edit_handlers import FieldPanel, StreamFieldPanel
from wagtail.core import blocks as wagtail_blocks
from wagtail.core.fields import StreamField
from wagtail.core.models import Page

from wagtailinventory.tests.testapp import blocks


class NoStreamFieldsPage(Page):
    content = models.TextField()

    content_panels = Page.content_panels + [
        FieldPanel("content"),
    ]


class SingleStreamFieldPage(Page):
    content = StreamField(
        [
            ("atom", blocks.Atom()),
            ("molecule", blocks.Molecule()),
            ("organism", blocks.Organism()),
        ],
        blank=True,
    )

    content_panels = Page.content_panels + [
        StreamFieldPanel("content"),
    ]


class MultipleStreamFieldsPage(Page):
    first = StreamField(
        [
            ("atom", blocks.Atom()),
            ("molecule", blocks.Molecule()),
            ("organism", blocks.Organism()),
        ]
    )
    second = StreamField(
        [
            ("atom", blocks.Atom()),
            ("molecule", blocks.Molecule()),
            ("organism", blocks.Organism()),
        ]
    )

    content_panels = Page.content_panels + [
        StreamFieldPanel("first"),
        StreamFieldPanel("second"),
    ]


class NestedStreamBlockPage(Page):
    content = StreamField(
        [
            (
                "streamblock",
                wagtail_blocks.StreamBlock(
                    [
                        ("text", wagtail_blocks.CharBlock()),
                        ("atom", blocks.Atom()),
                    ]
                ),
            ),
        ]
    )

    content_panels = Page.content_panels + [
        StreamFieldPanel("content"),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0032_add_bulk_delete_page_permission'),
    ]

    operations = [
        migrations.CreateModel(
            name='PageBlock',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('block', models.CharField(db_index=True, max_length=255)),
                ('page', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='page_blocks', to='wagtailcore.Page')),
            ],
        ),
    ]



from django import forms
from django.db.models.fields import BLANK_CHOICE_DASH
from django.forms import formset_factory

from wagtail.core.models import Page

from wagtailinventory.models import PageBlock


class PageBlockQueryForm(forms.Form):
    INCLUDES_BLOCK = "includes"
    EXCLUDES_BLOCK = "excludes"

    block = forms.ChoiceField(
        choices=(),
        required=False,
        label="Block type",
    )

    has = forms.ChoiceField(
        choices=((c, c) for c in (INCLUDES_BLOCK, EXCLUDES_BLOCK)),
        label=None,
    )

    def __init__(self, *args, **kwargs):
        super(PageBlockQueryForm, self).__init__(*args, **kwargs)
        blocks = (
            PageBlock.objects.values_list("block", flat=True)
            .distinct()
            .order_by("block")
        )

        block_choices = BLANK_CHOICE_DASH + [(b, b) for b in blocks]
        self.fields["block"].choices = block_choices


class BasePageBlockQueryFormSet(forms.BaseFormSet):
    def get_query(self):
        qs = Page.objects.all()

        for form in self.forms:
            if not form.is_valid():
                continue

            form_block = form.cleaned_data.get("block")

            if not form_block:
                continue

            condition = {"page_blocks__block": form_block}

            if form.cleaned_data["has"] == form.INCLUDES_BLOCK:
                qs = qs.filter(**condition)
            else:
                qs = qs.exclude(**condition)

        return qs


PageBlockQueryFormSet = formset_factory(
    PageBlockQueryForm, formset=BasePageBlockQueryFormSet, extra=3, max_num=3
)



from django.core.management import BaseCommand

from wagtail.core.models import Page

from tqdm import tqdm

from wagtailinventory.helpers import (
    create_page_inventory,
    delete_page_inventory,
)


class Command(BaseCommand):
    def handle(self, *args, **options):
        delete_page_inventory()

        pages = Page.objects.all()

        if options.get("verbosity"):  # pragma: no cover
            pages = tqdm(pages)

        for page in pages:
            create_page_inventory(page)



from django.db import migrations, models
from django.db.models import Min, Count


def remove_duplicates(apps, schema_editor):
    PageBlock = apps.get_model('wagtailinventory', 'PageBlock')

    duplicate_pks_to_keep = (
        PageBlock.objects
            .values('page', 'block')
            .annotate(Min('pk'), count=Count('pk'))
            .filter(count__gt=1)
            .values_list('pk__min', flat=True)
    )

    duplicates_to_keep = PageBlock.objects.in_bulk(duplicate_pks_to_keep)

    for duplicate_to_keep in duplicates_to_keep.values():
        duplicates_to_delete = PageBlock.objects.filter(
            page=duplicate_to_keep.page,
            block=duplicate_to_keep.block
        ).exclude(pk=duplicate_to_keep.pk)

        duplicates_to_delete.delete()


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailinventory', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(remove_duplicates, migrations.RunPython.noop),
        migrations.AddConstraint(
            model_name='pageblock',
            constraint=models.UniqueConstraint(fields=('page', 'block'), name='unique_page_block'),
        ),
    ]



from django.conf import settings

from wagtail.admin import urls as wagtailadmin_urls
from wagtail.core import urls as wagtailcore_urls
from wagtail.documents import urls as wagtaildocs_urls


try:
    from django.urls import include, re_path
except ImportError:
    from django.conf.urls import include
    from django.conf.urls import url as re_path


urlpatterns = [
    re_path(r"^admin/", include(wagtailadmin_urls)),
    re_path(r"^documents/", include(wagtaildocs_urls)),
    re_path(r"", include(wagtailcore_urls)),
]


if settings.DEBUG:
    from django.conf.urls.static import static
    from django.contrib.staticfiles.urls import staticfiles_urlpatterns

    urlpatterns += staticfiles_urlpatterns()
    urlpatterns += static(
        settings.MEDIA_URL, document_root=settings.MEDIA_ROOT
    )



from urllib.parse import urlencode

from django.core.management import call_command
from django.test import TestCase
from django.urls import reverse

from wagtail.core.models import Page
from wagtail.tests.utils import WagtailTestUtils


class SearchViewTests(WagtailTestUtils, TestCase):
    fixtures = ["test_blocks.json"]

    def setUp(self):
        self.login()

    def get(self, params=None):
        path = reverse("wagtailinventory:search")

        if params:
            path += "?" + urlencode(params)

        return self.client.get(path)

    def test_empty_query_returns_200_and_all_pages_ordered_by_title(self):
        response = self.get()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.context["pages"].object_list,
            list(Page.objects.order_by("title")),
        )

    def test_get_bad_formset_query_returns_400(self):
        response = self.get(
            {
                "form-TOTAL_FORMS": 1,
                "form-INITIAL_FORMS": 0,
                "form-0-has": "invalid",
            }
        )
        self.assertEqual(response.status_code, 400)

    def test_valid_query_returns_200_and_only_appropriate_pages(self):
        call_command("block_inventory", verbosity=0)

        response = self.get(
            {
                "form-TOTAL_FORMS": 1,
                "form-INITIAL_FORMS": 0,
                "form-0-has": "includes",
                "form-0-block": "wagtailinventory.tests.testapp.blocks.Organism",  # noqa
            }
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.context["pages"].object_list,
            [Page.objects.get(slug="multiple-streamfields-page")],
        )



from setuptools import find_packages, setup


install_requires = [
    "tqdm>=4.15.0,<5",
    "wagtail>=2.15,<3",
]


testing_extras = [
    "coverage>=3.7.0",
    "mock>=1.0.0",
]

setup(
    name="wagtail-inventory",
    url="https://github.com/cfpb/wagtail-inventory",
    author="CFPB",
    author_email="tech@cfpb.gov",
    description="Lookup Wagtail pages by block content",
    long_description=open("README.rst", "r", encoding="utf-8").read(),
    license="CCO",
    version="1.5",
    include_package_data=True,
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=install_requires,
    extras_require={"testing": testing_extras},
    classifiers=[
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Wagtail",
        "Framework :: Wagtail :: 2",
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        "License :: Public Domain",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)



from django.apps import AppConfig


class WagtailInventoryAppConfig(AppConfig):
    name = "wagtailinventory"
    verbose_name = "Wagtail Inventory"
    default_auto_field = "django.db.models.BigAutoField"
